-- $Id: CombiningMonads.hs 78 2008-01-05 21:36:03Z inf6254 $
-- -----------------------------------------------------------------------------
module CombiningMonads where

import Evaluator
    ( Term(..)
    , answer, wrong
    , Exc(..)
    , Exception
    )

class Monad m => ExMonad m where
    raise               :: Exception -> m a

class Monad m => StMonad m where
    tick                :: m ()

class Monad m => ShowMonad m where
    showMonad           :: m String -> String

class Transformer t where
    promote             :: Monad m => m a -> t m a

-- -----------------------------------------------------------------------------
-- Evaluierer mit Exception und State
-- -----------------------------------------------------------------------------
eval                    :: (ExMonad m, StMonad m) => Term -> m Int
eval (Con x)            =  return x
eval (Div t u)          =  do
                           x <- eval t
                           y <- eval u
                           tick
                           if y == 0
                              then raise "division by zero"
                              else return (x `div` y)

mapMonad                :: Monad m => (a -> b) -> m a -> m b
mapMonad f p            =  do {x <- p; return (f x)}

-- -----------------------------------------------------------------------------
-- Combining Exceptions
-- -----------------------------------------------------------------------------
newtype EXC m a         =  MkEXC (m (Exc a))

recover                 :: EXC m a -> m (Exc a)
recover (MkEXC g)       =  g

instance Monad m => Monad (EXC m) where
    return x            =  MkEXC (return (Return x))
    p >>= q             =  MkEXC (recover p >>= r)
                           where
                           r (Raise e)  = return (Raise e)
                           r (Return x) = recover (q x)

instance Monad m => ExMonad (EXC m) where
    raise e             =  MkEXC (return (Raise e))

instance Transformer EXC where
    promote g           =  MkEXC (do {x <- g; return (Return x)})

instance StMonad m => StMonad (EXC m) where
    tick                =  promote tick

instance ShowMonad m => ShowMonad (EXC m) where
    showMonad p         =  showMonad (recover p >>= q)
                           where
                           q (Raise e)  = return ("exception: " ++ e)
                           q (Return x) = return x

instance (ShowMonad m, Show a) => Show (EXC m a) where
    show                =  showMonad . mapMonad show

-- -----------------------------------------------------------------------------
-- Combining State
-- -----------------------------------------------------------------------------
newtype STT m a         =  MkSTT (State -> m (a, State))
type State              =  Int

apply                   :: STT m a -> State -> m (a, State)
apply (MkSTT f)         =  f

instance Monad m => Monad (STT m) where
    return x            =  MkSTT f where f s = return (x,s)
    p >>= q             =  MkSTT f
                           where
                           f s = do
                                 (x,s') <- apply p s
                                 apply (q x) s'

instance Monad m => StMonad (STT m) where
    tick                =  MkSTT f
                           where
                           f s = return ((),s+1)

instance ShowMonad m => ShowMonad (STT m) where
    showMonad p         =  showMonad q
                           where
                           q = do 
                               (x,s) <- apply p 0
                               return ("value: " ++ x ++ 
                                 ", count: " ++ show s)

instance (ShowMonad m, Show a) => Show (STT m a) where
    show                =  showMonad . mapMonad show

instance Transformer STT where
    promote g           =  MkSTT f
                           where
                           f s = do {x <- g; return (x,s)}

instance ExMonad m => ExMonad (STT m) where
    raise e             =  promote (raise e)
    
-- -----------------------------------------------------------------------------
-- Putting together
-- -----------------------------------------------------------------------------
newtype Id a            =  MkId a

instance Monad Id where
    return x            =  MkId x
    (MkId x) >>= q      =  q x

instance ShowMonad Id where
    showMonad (MkId cs) =  cs

evalExSt                :: Term -> EXC (STT Id) Int
evalExSt                =  eval

evalStEx                :: Term -> STT (EXC Id) Int
evalStEx                =  eval